package v6.name.panitz.wip;

import java.io.FileOutputStream;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

import v6.name.panitz.wip.Tree.ConDef;
import v6.name.panitz.wip.Tree.FieldSelect;
import v6.name.panitz.wip.Tree.FunCall;
import v6.name.panitz.wip.Tree.FunDef;
import v6.name.panitz.wip.Tree.IntLiteral;
import v6.name.panitz.wip.Tree.OpExpr;
import v6.name.panitz.wip.Tree.Param;
import v6.name.panitz.wip.Tree.Program;
import v6.name.panitz.wip.Tree.Variable;

public class GenJBC implements Visitor<Void> {
	ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES|ClassWriter.COMPUTE_MAXS);
	MethodVisitor mv;
	
	Map<String, Tree.FunDef> funs;
	Map<String, Integer> vars;
	Map<String, Tree.ConDef> cons;

	public GenJBC(Tree t) {
	}
	
	@Override
	public Void visit(IntLiteral il) {
		mv.visitLdcInsn(il.n);
		return null;
	}

	@Override
	public Void visit(OpExpr oe) {
		oe.lhs.welcome(this);
		oe.rhs.welcome(this);
		switch (oe.op) {
		case mult:
			mv.visitInsn(Opcodes.IMUL);			
			break;

		default:
			mv.visitInsn(Opcodes.IADD);			
			break;
		}
		return null;
	}
	@Override
	public Void visit(FunDef funDef) {
		var args = "";
		for (var a:funDef.args)args+=a.getJBCType();
		mv=cw.visitMethod(Opcodes.ACC_PUBLIC|Opcodes.ACC_STATIC, funDef.name,"("+args+")I", null, null);

		vars = new HashMap<>();
		var i = 0;
		for (var a:funDef.args)vars.put(a.name, i++);
		funDef.body.welcome(this);
		mv.visitInsn(Opcodes.IRETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();
		return null;
	}
	@Override
	public Void visit(Program program) {
		program.cons.forEach(con -> con.welcome(this));	
		
		
		cw.visit(Opcodes.V1_8, Opcodes.ACC_PUBLIC+Opcodes.ACC_SUPER, "Main", null, "java/lang/Object", null);
		cw.visitSource("Main.wip", null);

		funs = new HashMap<>();
		program.funs.forEach(fun->funs.put(fun.name, fun));
		cons = new HashMap<>();
		program.cons.forEach(con->cons.put(con.name, con));
		
		
		program.funs.forEach(fun->fun.welcome(this));
		
		mv=cw.visitMethod(Opcodes.ACC_PUBLIC|Opcodes.ACC_STATIC, "start","()I", null, null);
		program.main.welcome(this);
		mv.visitInsn(Opcodes.IRETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();
		cw.visitEnd();
		try {
			var out = new FileOutputStream("Main.class");
			out.write(cw.toByteArray());
			out.close();
		}catch (Exception e) {
			
		}
		return null;
	}
	@Override
	public Void visit(Variable variable) {
		mv.visitVarInsn(Opcodes.ILOAD, vars.get(variable.name));
		return null;
	}
	@Override
	public Void visit(FunCall funCall) {
		var fun = funs.get(funCall.name);
		if (fun == null) {
			var con = cons.get(funCall.name);
			if (con==null) {
				throw new RuntimeException("unkown function name: "+funCall.name);
			}
	        mv.visitTypeInsn(Opcodes.NEW, con.name);
	        mv.visitInsn(Opcodes.DUP);
			funCall.params.forEach(param->param.welcome(this));
			String signature = "(";
			var i = 1;
			for (var arg : con.args) {
				signature += arg.getJBCType();
				i++;
			}

			mv.visitMethodInsn(Opcodes.INVOKESPECIAL, con.name, "<init>", signature+")V", false);
			
		}else {
		
			funCall.params.forEach(param->param.welcome(this));
			var args = "";
			for (var a:fun.args) args+="I";
			mv.visitMethodInsn(Opcodes.INVOKESTATIC, "Main", funCall.name, "("+args+")I",false);
		}
		return null;
	}

	@Override
	public Void visit(ConDef conDef) {
		cw.visit(Opcodes.V1_8, Opcodes.ACC_PUBLIC+Opcodes.ACC_SUPER, conDef.name, null, "java/lang/Object", null);
		cw.visitSource("Main.wip", null);
		for (var a :conDef.args) {
			cw.visitField(Opcodes.ACC_PUBLIC, a.name, a.getJBCType(), null, null);
		}
		
		String signature = "(";
		var i = 1;
		for (var arg : conDef.args) {
			signature += arg.getJBCType();
			i++;
		}
		MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "<init>", signature + ")V", null, null);
		mv.visitVarInsn(Opcodes.ALOAD, 0);
		mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "java/lang/Object", "<init>", "()V", false);
		i = 1;
		for (var arg : conDef.args) {	
			mv.visitVarInsn(Opcodes.ALOAD, 0);
			if (arg.typ.equals("int")) {
				mv.visitVarInsn(Opcodes.ILOAD, i);					
			}else {
				mv.visitVarInsn(Opcodes.ALOAD, i);									
			}
			mv.visitFieldInsn(Opcodes.PUTFIELD, conDef.name, arg.name, arg.getJBCType());
			i++;
		}
		
		mv.visitInsn(Opcodes.RETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();
		generateToString(cw, conDef);
		
		cw.visitEnd();
		try {
			var out = new FileOutputStream(conDef.name+".class");
			out.write(cw.toByteArray());
			out.close();
		}catch (Exception e) {
			
		}

		return null;
	}
	
	private void generateToString(ClassWriter cv, ConDef conDef) {
		MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "toString", "()Ljava/lang/String;", null, null);
		mv.visitTypeInsn(Opcodes.NEW, "java/lang/StringBuffer");
		mv.visitInsn(Opcodes.DUP);
		mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "java/lang/StringBuffer", "<init>", "()V", false);
		mv.visitVarInsn(Opcodes.ASTORE, 1);
		mv.visitVarInsn(Opcodes.ALOAD, 1);
		mv.visitLdcInsn(conDef.name);
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
				"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
		mv.visitLdcInsn("(");
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
				"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
		int i = 1;
		for (var arg : conDef.args) {
			if (i > 1) {
				mv.visitLdcInsn(", ");
				mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
						"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
			}
			boolean prim = arg.typ.equals("int");
			mv.visitVarInsn(Opcodes.ALOAD, 0);
	//		String typ = prim ? "I" : "Ljava/lang/Object;";
			mv.visitFieldInsn(Opcodes.GETFIELD, conDef.name, arg.name, arg.getJBCType());
			mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
					"(" + arg.getJBCType() + ")Ljava/lang/StringBuffer;", false);

			i++;
		}
		mv.visitLdcInsn(")");
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
				"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "toString", "()Ljava/lang/String;", false);
		mv.visitInsn(Opcodes.ARETURN);
		mv.visitMaxs(1, 1);
		mv.visitEnd();

	}
	

	@Override
	public Void visit(Param param) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Void visit(FieldSelect fieldSelect) {
		// TODO Auto-generated method stub
		return null;
	}

}
